<?php
if (!defined('MEDIAWIKI')) die("MediaWiki extensions cannot be run directly.");
if (!defined('WikiDB')) {
	die("The WikiDB extension has not been set up correctly.<br>'"
		. basename(__FILE__) . "' should not be included directly.<br>Instead, "
		. "add the following to LocalSettings.php: "
		. "<tt><b>include_once(\"WikiDB/WikiDB.php\");</b></tt>");
}
/**
 * General-purpose utility functions for WikiDB.
 * These may be useful in other applications too.
 *
 * To avoid conflicts with other extensions (particularly my own) I either check
 * that the function hasn't already been defined prior to defining it, or else
 * I've prefixed the function name with "WikiDB_", to make the name unique.
 *
 * @package MediaWiki
 * @subpackage Extensions
 *
 * @author Mark Clements <mclements at kennel17 dot co dot uk>
 * @copyright Copyright  2006-2014, Mark Clements
 * @license http://creativecommons.org/licenses/by-sa/2.5/ cc-by-sa 2.5 or later
 * @version $Rev: 856 $
 */

WikiDB_RegisterRevision('$Rev: 856 $');

// WikiDB_IsNumeric()
// Returns true if the supplied value is a valid number.
// Should be used instead of PHP's is_numeric() function as it is a lot stricter.
// The PHP function will return true for strings like "23 Skidoo" or "", which we
// don't want!
	function WikiDB_IsNumeric($Value) {
		if (preg_match('/^-?([0-9]+(\.[0-9]*)?|[0-9]*\.[0-9]+)$/', $Value))
			return true;
		else
			return false;
	}

// WikiDB_IsInteger()
// Same as IsNumeric() but does not allow non-integer values.  Note that
// "5.0" is currently not allowed, but "5" is.
// TODO: Should we allow non-integer representations of integers?
	function WikiDB_IsInteger($Value) {
		if (preg_match('/^-?[0-9]+$/', $Value))
			return true;
		else
			return false;
	}

// WikiDB_StrToNumber()
// Converts a string to a number, which is returned as a float().
// Zero is returned for non-numeric strings.
	function WikiDB_StrToNumber($Value) {
		if (WikiDB_IsNumeric($Value))
			return floatval($Value);
		else
			return 0;
	}

// WikiDB_StrToInteger()
// Converts a string to an integer value. Zero is returned for non-numeric strings.
// Non-integer values are rounded to the nearest whole number.
// Note: Despite being an integer, we actually return a float value, so that we
//		 are not restricted to the platform's integer precision.
	function WikiDB_StrToInteger($Value) {
		return round(WikiDB_StrToNumber($Value));
	}

// WikiDB_FormatFloat()
// Formats a float value as a string, with trailing zeros removed and a minimum of
// 1 digit before the decimal point, according to the current user's language.
	function WikiDB_FormatFloat($Value) {
		global $wgLang;

		$Value = sprintf("%1f", $Value);
		$Value = preg_replace("/(\.\d+)0+$/U", "$1", $Value);
		
		return $wgLang->formatNum($Value);
	}

// WikiDB_FormatInteger()
// Formats an integer as a string, according to the user's language.
	function WikiDB_FormatInteger($Value) {
		global $wgLang;

		$Value = sprintf("%1.0f", $Value);
		return $wgLang->formatNum($Value);
	}

// WikiDB_NumberToSortableString()
// Takes a number and formats it into a string-represenation suitable for sorting
// via a string sort.  If the input is not a valid number, then the function
// returns an empty string.
	function WikiDB_NumberToSortableString($Value) {
	// If value is blank, or is not a valid number, then return an empty string.
		if ($Value == "" || !WikiDB_IsNumeric($Value))
			return "";

	// Split the number on the decimal separator, to get the two components.
		$Parts = explode(".", $Value);

	// The part before the decimal point can simply be converted to an integer to
	// remove any leading zeros and to ensure that blank values are translated to
	// a zero (so that ".1" and "0.1" are considered equal).
		$Parts[0] = intval($Parts[0]);

	// The part after the decimal point needs to be trimmed of trailing zeros,
	// but any leading zeros need to be preserved, so we need to use a string
	// manipulation rather than using intval().
	// Also, if there is no second part, we normalise this to '0', so that
	// "1" and "1.0" are considered equal.
		if (isset($Parts[1])) {
			$Parts[1] = rtrim($Parts[1], "0");
			if (IsBlank($Parts[1]))
				$Parts[1] = "0";
		}
		else {
			$Parts[1] = 0;
		}

	// Set the sign byte.  We use two values that we know will be in the
	// correct sort order, regardless of the character set.
		if ($Parts[0] < 0) {
			$Parts[0] = abs($Parts[0]);
			$Sign = "n";
		}
		else
			$Sign = "p";

	// Field size is 255. 3 bytes are required for the initial sort index,
	// the sign and the decimal point. This leaves 126 bytes of padding for
	// each part. We use a space at the start, to ensure numbers are sorted
	// _before_ strings.  We don't bother to pad the numbers after the decimal
	// point, as they are left-aligned and so the padding makes no difference.
		return sprintf(" %s%126d.%s", $Sign, $Parts[0], $Parts[1]);
	}

// pWikiDB_ExpandVariables()
// Simple function to expand variables in a string, e.g. {{{2}}} is replaced with
// the value in element 2 of the array.  Used when we expand template arguments when
// rendering a <repeat> tag.
// TODO - this is VERY crude at the moment, and may well fail in certain edge cases
// (or even, certain 'non-edge' cases!)
	function pWikiDB_ExpandVariables($Input, $Data) {
		if (is_array($Data) && count($Data) > 0) {
		// Substitute variables.  Note that we need to replace $ with \$ within
		// $Value, to stop it being interpreted as a PHP variable reference.
			foreach ($Data as $Key => $Value) {
				$Input = preg_replace("/{{{" . preg_quote($Key, "/") . "}}}/",
									  str_replace('$', '\$', $Value),
									  $Input);
			}
		}
		return $Input;
	}

// WikiDB_Parse()
// Wrapper function to simplify text parsing.
// We use a local instance of the parser class, so that we don't get into a
// recursive nightmare (e.g. if a <repeat> tag contains a <repeat> tag).
// All tag hooks are ignored (as they are being dealt with already), however
// parser function hooks are parsed as expected.
	function WikiDB_Parse($Input, &$Parser, $LineStart = true) {
		global $wgParser;
		static $LocalParser;
		
		if (!isset($LocalParser)) {
			$LocalParser = new Parser;
		
		// Later versions of MW (tested on 1.16) automatically add the necessary parser 
		// function hooks to a new Parser instance, however earlier versions (tested
		// on 1.6) don't.  If they are not present, copy them from the global
		// $wgParser object.
			if (count($LocalParser->mFunctionHooks) == 0)
				$LocalParser->mFunctionHooks = $wgParser->mFunctionHooks;
		}

		$Output = $LocalParser->parse($Input, $Parser->mTitle, $Parser->mOptions,
									  $LineStart, true);
		return $Output->getText();
	}

// WikiDB_IsValidTable()
// Wrapper for WikiDB_Table::IsValidTable().  That function can only be called once
// you've instantiated a table object, but often we need to check whether the raw
// title we have received refers to a valid table without having to create a table
// object for it.  This function allows us to do that.
// Arguments are the same as for the WikiDB_Table constructor, so are quite flexible.
// Note that previously we allowed the class method to be called both statically
// and non-statically, but this throws E_STRICT warnings, due to a stupid design
// choice by the PHP authors.  Oh well.  We write more complicated code to keep the
// error logs happy. :-(
	function WikiDB_IsValidTable($Title, $DefaultNamespace = NULL) {
		$Table = new WikiDB_Table($Title, $DefaultNamespace);
		return $Table->IsValidTable();
	}

// WikiDB_MakeTitle()
// $Title may either be a title object or a string representing the name of a
// page.  If $Title is a string, but doesn't include a namespace, then
// $DefaultNamespace is used if specified.  If no default is specified then 
// WikiDB::GetDefaultTableNamespace() is used to fill in a sensible default.
	function WikiDB_MakeTitle($Title, $DefaultNamespace = NULL) {
	// If $Title is already a Title object, just return it.
	// TODO: Currently we assume that if it is an object, it is a Title object.
	//		 Is this sensible?
		if (is_object($Title)) {
			return $Title;
		}
	// Otherwise, it is a text string representing a page title.
	// In this case we create a new title object from that string, using the
	// default table namespace if necessary.
		else {
		// If no default namespace is specified, use the standard default.
			if ($DefaultNamespace === NULL)
				$DefaultNamespace = WikiDB::GetDefaultTableNamespace();

		// TODO: Should this use Title::newFromDBKey() instead?  If $Title contains
		//		 text it *should* always be a DB key, but it seems that the
		//		 transformations involved in newFromText() won't affect strings that
		//		 are already transformed, therefore it is safe (if marginally slower)
		//		 to pass it through there anyway - and that avoids problems if we
		//		 get passed in the actual title, rather than the DB key, for any
		//		 reason.
			return Title::newFromText($Title, $DefaultNamespace);
		}
	}

// GetCanonicalNamespaceName()
// Returns the canonical name of the specified namespace.  I've written a wrapper
// function, as this isn't as simple as I'd like!
	function WikiDB_GetCanonicalNamespaceName($NS) {
		global $wgCanonicalNamespaceNames;

	// If the namespace exists in $wgCanonicalNamespaceNames then get its canoncial
	// name using the getCanonicalName() function of the MWNamespace class.
	// Originally, this class was called Namespace, but this became a reserved word in
	// PHP 5.3 and so in MW 1.13 it was renamed to MWNamespace.  Namespace remained
	// as an alias until it was finally removed altogether in MW 1.17.
		if (isset($wgCanonicalNamespaceNames[$NS])) {
		// In MW >= 1.13, the class we need is called MWNamespace, so if the MWNamespace
		// class exists, use it.
			if (class_exists("MWNamespace")) {
				$NSName = MWNamespace::getCanonicalName($NS);
			}
		// If MWNamespace is unavailable, then we are using a verions of MW < 1.13.
		// In these earlier versions, the class was called Namespace, so we use that
		// class instead.  However, we can't simply call this class in the normal
		// fashion, as it would cause parse errors if running on PHP >= 5.3, therefore 
		// we need to use eval() to run the necessary code, as this won't be parsed
		// until it is actually executed.
			else {
				$PHPCode = 'return Namespace::getCanonicalName(' . intval($NS) . ');';
				$NSName = eval($PHPCode);
			}
		}
	// If it is not defined, but $NS = 0, then it is the main namespace.  This doesn't
	// have an actual name, but we still need to report it.  We therefore use 'main' for
	// this.
	// TODO: Can this be internationalised?
		elseif ($NS == 0) {
			$NSName = "main";
		}
	// Otherwise, it is an undefined namespace.
	// TODO: Can this be internationalised?
		else {
			$NSName = "undefined";
		}
		
		return $NSName;
	}

// WikiDB_GetAction()
// Returns the current page action.  We wrap it into a function because the method
// for dealing with this has changed between MW versions.  This ensures we can
// easily maintain compatibility.
	function WikiDB_GetAction() {
		global $action, $mediaWiki;

	// The $mediaWiki object has existed since very early versions (was present in
	// MW 1.6, for example).  However the getAction() method wasn't added until
	// MW 1.18.  Therefore we need to check whether this function exists before
	// calling it, to ensure compatibility with older versions.
	// If it doesn't exist, we use the global $action variable instead, which is
	// the old way of getting this same information.
		if (is_callable(array($mediaWiki, "getAction")))
			return $mediaWiki->getAction();
		else
			return $action;
	}

// WikiDB_GetNamespaceIDs()
// Returns an array of namespace names => namespace IDs.  This is effectively a
// wrapper for $wgContLang->getNamespaceIds(), but in earlier versions of MediaWiki,
// this function doesn't exist, so in that situation we create the array ourselves
// using the same logic as is used in later MW versions.
// Required for MW 1.7 support, but possibly for later versions too as I'm not sure
// when the function was added.
	function WikiDB_GetNamespaceIDs() {
		global $wgContLang, $wgNamespaceAliases;
		static $NamespaceIDs;

		if (method_exists($wgContLang, "getNamespaceIds"))
			return $wgContLang->getNamespaceIds();
		else {
			if (is_null($NamespaceIDs)) {
				# Put namespace names and aliases into a hashtable.
				# If this is too slow, then we should arrange it so that it is done
				# before caching. The catch is that at pre-cache time, the above
				# class-specific fixup hasn't been done.
				$NamespaceIDs = array();
				foreach ($wgContLang->getNamespaces() as $Index => $Name) {
					$NamespaceIDs[$wgContLang->lc($Name)] = $Index;
				}
				if (method_exists($wgContLang, "getNamespaceAliases")) {
					foreach ($wgContLang->getNamespaceAliases() as $Name => $Index) {
						$NamespaceIDs[$wgContLang->lc($Name)] = $Index;
					}
				}
				if ($wgNamespaceAliases) {
					foreach ($wgNamespaceAliases as $Name => $Index) {
						$NamespaceIDs[$wgContLang->lc($Name)] = $Index;
					}
				}
			}

			return $NamespaceIDs;
		}
	}

// IsBlank()
// A staggeringly useful function when you don't care much about the type of blank.
// We wrap in a function_exists() check so that it plays nicely when used with my
// other extensions, some of which also include this function.
if (!function_exists("IsBlank")) {
	function IsBlank($Input) {
	// is_null() returns true if $Input is unset or null
		if (is_null($Input))
			return true;
		elseif ($Input === "")
			return true;
		elseif (is_array($Input) && count($Input) == 0)
			return true; 
		else
			return false;
	}
}
